
# This file was generated by 'versioneer.py' (0.17) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2017-05-14T11:54:35+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "4f0227ca58868e81cbc16efb9940ed0ea5f9a5d1",
 "version": "v1.4.4"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
